/*
 * Design overview:
 *
 * stun implements a point to point encrypted tunnel.  It supports
 * layer 2 (TAP) and layer 3 (TUN) tunnels.  TCP is the only supported
 * transport.
 *
 * All communication is encrypted using a pre-shared symmetric key.
 * aes-256-gcm was chosen because it is fast and does authentication.
 * The key is derived with PBKDF2.
 *
 * stun is a client-server design.  There can only be a single client
 * connected to the server at any time.  Routing between clients is handled
 * by the networking stack on the server side.
 *
 * When a client connects there is a mutual challenge-response phase
 * as shown below:
 *
 * t0: server challenges client
 * t1: client responds to server's challenge
 * t2: client challenges server
 * t3: server responds to client's challenge
 *
 * The challenge-response algorithm is as follows:
 *
 * The challenge is a randomly generated 64-bit integer encrypted
 * with the pre-shared key and sent to the receiver.
 * The receiver decrypts it, adds 1, encrypts it and sends it back to
 * the sender.  The sender verifies the response.
 * This algorithm is the same as what Kerberos uses for authentication.
 *
 * The stun packet format is shown below:
 *
 * [TAG] [IV] [PAYLOAD LENGTH] [PAYLOAD]
 *
 * The tag is used for authentication and it is 16 octets.
 * The iv is randomly generated for each packet and it is 12 octets.
 * The payload length is 2 octets.
 */

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>
#if defined(__linux__)
#include <linux/if_tun.h>
#include <bsd/stdlib.h>
#elif defined(__OpenBSD__) || defined(__FreeBSD__)
#include <net/if_tun.h>
#elif defined(__DragonFly__)
#include <net/tun/if_tun.h>
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <poll.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <openssl/aes.h>
#include <openssl/evp.h>

#include "arg.h"

#if defined(__linux__) || defined(__DragonFly__) || defined(__FreeBSD__)
#undef explicit_bzero
#define explicit_bzero bzero
#endif

#define NOPRIVUSER	"nobody"
#define RCVTIMEO	250 /* in milliseconds */
#define RECONNECTTIMEO	60  /* in seconds */
#define TAGLEN		16
#define IVLEN		12
#define HDRLEN		2
#define MAXPAYLOADLEN	1428
#define MAXPKTLEN	(TAGLEN + IVLEN + HDRLEN + MAXPAYLOADLEN)
#define BADPKT		0x1000

enum {
	TUNDEV,
	TAPDEV
};

EVP_CIPHER_CTX ectx, dctx;
unsigned char aeskey[EVP_MAX_KEY_LENGTH];
char *argv0;
char *bindaddr;
char *host;
char *port = "12080";
int devtype = TUNDEV;
int debug;
int foreground;
int sflag;

void
logmsg(int priority, char *msg, va_list ap)
{
	if (foreground) {
		fputs("stun: ", stderr);
		vfprintf(stderr, msg, ap);
		fputc('\n', stderr);
	} else {
		vsyslog(priority, msg, ap);
	}
}

void
logdbg(char *msg, ...)
{
	va_list ap;

	if (debug) {
		va_start(ap, msg);
		logmsg(LOG_DAEMON | LOG_DEBUG, msg, ap);
		va_end(ap);
	}
}

void
logwarn(char *msg, ...)
{
	va_list ap;

	va_start(ap, msg);
	logmsg(LOG_DAEMON | LOG_WARNING, msg, ap);
	va_end(ap);
}

void
logerr(char *msg, ...)
{
	va_list ap;

	va_start(ap, msg);
	logmsg(LOG_DAEMON | LOG_ERR, msg, ap);
	va_end(ap);
	exit(1);
}

void
pack16(unsigned char *buf, uint16_t n)
{
	buf[0] = n >> 8 & 0xff;
	buf[1] = n & 0xff;
}

uint16_t
unpack16(unsigned char *buf)
{
	return buf[0] << 8 | buf[1];
}

void
pack64(unsigned char *buf, uint64_t n)
{
	buf[0] = n >> 56 & 0xff;
	buf[1] = n >> 48 & 0xff;
	buf[2] = n >> 40 & 0xff;
	buf[3] = n >> 32 & 0xff;
	buf[4] = n >> 24 & 0xff;
	buf[5] = n >> 16 & 0xff;
	buf[6] = n >> 8 & 0xff;
	buf[7] = n & 0xff;
}

uint64_t
unpack64(unsigned char *buf)
{
	return (uint64_t)buf[0] << 56 |
	       (uint64_t)buf[1] << 48 |
	       (uint64_t)buf[2] << 40 |
	       (uint64_t)buf[3] << 32 |
	       (uint64_t)buf[4] << 24 |
	       (uint64_t)buf[5] << 16 |
	       (uint64_t)buf[6] << 8  |
	       (uint64_t)buf[7];
}

int
writeall(int fd, void *buf, int len)
{
	unsigned char *p = buf;
	int n, total = 0;

	while (len > 0) {
		n = write(fd, p + total, len);
		if (n <= 0)
			break;
		total += n;
		len -= n;
	}
	return total;
}

int
readall(int fd, void *buf, int len)
{
	unsigned char *p = buf;
	int n, total = 0;

	while (len > 0) {
		n = read(fd, p + total, len);
		if (n <= 0)
			break;
		total += n;
		len -= n;
	}
	return total;
}

void
ms2tv(struct timeval *tv, long ms)
{
	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}

void
revokeprivs(void)
{
	struct passwd *pw;

	pw = getpwnam(NOPRIVUSER);
	if (!pw)
		logerr("no %s user", NOPRIVUSER);
	if (setgroups(1, &pw->pw_gid) ||
	    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) ||
	    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid))
		logerr("failed to revoke privs");
}

int
derivekey(char *pw, int pwlen)
{
	int nrounds = 100000;

	if (PKCS5_PBKDF2_HMAC_SHA1(pw, pwlen, NULL, 0, nrounds, 32, aeskey) != 1)
		logerr("PKCS5_PBKDF2_HMAC_SHA1 failed");
	return 0;
}

int
aesinit(EVP_CIPHER_CTX *ectx, EVP_CIPHER_CTX *dctx)
{
	EVP_CIPHER_CTX_init(ectx);
	EVP_CIPHER_CTX_init(dctx);
	return 0;
}

int
aesenc(EVP_CIPHER_CTX *ctx, unsigned char *ct, unsigned char *pt, int ptlen,
       unsigned char *key, unsigned char *iv, unsigned char *aad, int aadlen,
       unsigned char *tag, int taglen)
{
	int len, flen;

	/* initialize encryption operation with the given key and iv */
	if (EVP_EncryptInit_ex(ctx, EVP_aes_256_gcm(), NULL, key, iv) != 1)
		logerr("EVP_EncryptInit_ex failed");

	/* encrypt additional authentication data */
	if (EVP_EncryptUpdate(ctx, NULL, &len, aad, aadlen) != 1)
		logerr("EVP_EncryptUpdate failed");

	/* encrypt payload */
	if (EVP_EncryptUpdate(ctx, ct, &len, pt, ptlen) != 1)
		logerr("EVP_EncryptUpdate failed");

	/* finalize encryption */
	if (EVP_EncryptFinal_ex(ctx, ct + len, &flen) != 1)
		logerr("EVP_EncryptFinal_ex failed");

	/* get the tag */
	if (EVP_CIPHER_CTX_ctrl(ctx, EVP_CTRL_GCM_GET_TAG, TAGLEN, tag) != 1)
		logerr("EVP_CTRL_GCM_GET_TAG failed");

	return len + flen;
}

int
aesdec(EVP_CIPHER_CTX *ctx, unsigned char *pt, unsigned char *ct, int ctlen,
       unsigned char *key, unsigned char *iv, unsigned char *aad, int aadlen,
       unsigned char *tag, int taglen)
{
	int len, flen;

	/* initialize decryption operation with the given key and iv */
	if (EVP_DecryptInit_ex(ctx, EVP_aes_256_gcm(), NULL, key, iv) != 1)
		logerr("EVP_DecryptInit_ex failed");

	/* decrypt additional authentication data */
	if (EVP_DecryptUpdate(ctx, NULL, &len, aad, aadlen) != 1)
		logerr("EVP_DecryptUpdate failed");

	/* decrypt payload */
	if (EVP_DecryptUpdate(ctx, pt, &len, ct, ctlen) != 1)
		logerr("EVP_DecryptUpdate failed");

	/* set the tag */
	if (EVP_CIPHER_CTX_ctrl(ctx, EVP_CTRL_GCM_SET_TAG, TAGLEN, tag) != 1)
		logerr("EVP_CTRL_GCM_SET_TAG failed");

	/* finalize decryption and check if the tag matches */
	if (EVP_DecryptFinal_ex(ctx, pt + len, &flen) != 1)
		return -1;

	return len + flen;
}

#if defined(__linux__)
int
opendev(char *dev)
{
	struct ifreq ifr;
	int fd, s;

	fd = open("/dev/net/tun", O_RDWR);
	if (fd < 0)
		logerr("failed to open %s", "/dev/net/tun");

	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = (devtype == TUNDEV ? IFF_TUN : IFF_TAP) | IFF_NO_PI;
	strncpy(ifr.ifr_name, dev, IF_NAMESIZE);
	ifr.ifr_name[IF_NAMESIZE - 1] = '\0';
	if (ioctl(fd, TUNSETIFF, &ifr) < 0)
		logerr("failed to set TUNSETIFF on %s", dev);

	/* dummy socket so we can manipulate the params */
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		logerr("failed to create socket");
	if (devtype == TUNDEV)
		ifr.ifr_mtu = MAXPAYLOADLEN;
	else
		ifr.ifr_mtu = MAXPAYLOADLEN - 14; /* make some room for ethernet header */
	if (ioctl(s, SIOCSIFMTU, &ifr) < 0)
		logerr("failed to set MTU on %s", dev);
	close(s);

	return fd;
}

int
writedev(int fd, unsigned char *buf, int len)
{
	return write(fd, buf, len);
}

int
readdev(int fd, unsigned char *buf, int len)
{
	return read(fd, buf, len);
}
#elif defined(__OpenBSD__) || defined(__DragonFly__) || defined(__FreeBSD__)
int
opendev(char *dev)
{
	struct tuninfo ti;
	int fd;

	fd = open(dev, O_RDWR);
	if (fd < 0)
		logerr("failed to open %s", dev);
	if (ioctl(fd, TUNGIFINFO, &ti) < 0)
		logerr("failed to set TUNGIFINFO on %s", dev);
	if (devtype == TUNDEV)
		ti.mtu = MAXPAYLOADLEN;
	else
		ti.mtu = MAXPAYLOADLEN - 14; /* make some room for ethernet header */
	if (ioctl(fd, TUNSIFINFO, &ti) < 0)
		logerr("failed to set TUNSIFINFO on %s", dev);
	if (devtype == TUNDEV) {
#if defined(TUNSIFHEAD)
		int one = 1;
		if (ioctl(fd, TUNSIFHEAD, &one) < 0)
			logerr("failed to set TUNSIFHEAD on %s", dev);
#endif
	}
	return fd;
}

int
writedev(int fd, unsigned char *buf, int len)
{
	struct iovec iov[2];
	uint32_t type = htonl(AF_INET);
	int n;

	switch (devtype) {
	case TAPDEV:
		return write(fd, buf, len);
	case TUNDEV:
		iov[0].iov_base = &type;
		iov[0].iov_len = sizeof(type);
		iov[1].iov_base = buf;
		iov[1].iov_len = len;
		n = writev(fd, iov, 2);
		if (n > 0)
			n -= sizeof(type);
		break;
	}
	return n;
}

int
readdev(int fd, unsigned char *buf, int len)
{
	struct iovec iov[2];
	uint32_t type;
	int n;

	switch (devtype) {
	case TAPDEV:
		return read(fd, buf, len);
	case TUNDEV:
		iov[0].iov_base = &type;
		iov[0].iov_len = sizeof(type);
		iov[1].iov_base = buf;
		iov[1].iov_len = len;
		n = readv(fd, iov, 2);
		if (n > 0)
			n -= sizeof(type);
		break;
	}
	return n;
}
#endif

int
writenet(int fd, unsigned char *pt, int len)
{
	unsigned char ct[MAXPAYLOADLEN];
	unsigned char hdr[HDRLEN], iv[IVLEN], tag[TAGLEN];
	unsigned char pkt[MAXPKTLEN];
	int pktlen;

	pack16(hdr, len);
	arc4random_buf(iv, IVLEN);
	aesenc(&ectx, ct, pt, len, aeskey, iv, hdr, HDRLEN, tag, TAGLEN);
	memcpy(pkt, tag, TAGLEN);
	memcpy(&pkt[TAGLEN], iv, IVLEN);
	memcpy(&pkt[TAGLEN + IVLEN], hdr, HDRLEN);
	memcpy(&pkt[TAGLEN + IVLEN + HDRLEN], ct, len);
	pktlen = TAGLEN + IVLEN + HDRLEN + len;
	return writeall(fd, pkt, pktlen);
}

int
readnet(int fd, unsigned char *pt, int len)
{
	unsigned char ct[MAXPAYLOADLEN];
	unsigned char hdr[HDRLEN], iv[IVLEN], tag[TAGLEN];
	int n, ctlen;

#define CHECKERR(n) do { \
	if ((n) == 0) { \
		return 0; \
	} else if ((n) < 0) { \
		if (errno != EWOULDBLOCK) \
			return -1; \
		return BADPKT; \
	} \
} while (0)

	n = readall(fd, tag, TAGLEN);
	CHECKERR(n);

	n = readall(fd, iv, IVLEN);
	CHECKERR(n);

	n = readall(fd, hdr, HDRLEN);
	CHECKERR(n);

	if ((ctlen = unpack16(hdr)) > MAXPAYLOADLEN)
		ctlen = MAXPAYLOADLEN;

	n = readall(fd, ct, ctlen);
	CHECKERR(n);

	n = aesdec(&dctx, pt, ct, ctlen, aeskey, iv,
	           hdr, HDRLEN, tag, TAGLEN);
	if (n < 0)
		return BADPKT;
	return n;
}

int
challenge(int netfd)
{
	unsigned char buf[MAXPAYLOADLEN];
	struct pollfd pfd[1];
	uint64_t n, reply;
	int ret;

	arc4random_buf(&n, sizeof(uint64_t));
	pack64(buf, n);
	if (writenet(netfd, buf, sizeof(uint64_t)) <= 0)
		return -1;

	pfd[0].fd = netfd;
	pfd[0].events = POLLIN;
	ret = poll(pfd, 1, RCVTIMEO);
	if (ret < 0) {
		logwarn("poll failed");
		return -1;
	} else if (ret == 0) {
		logwarn("challenge-response timed out");
		return -1;
	}

	if (pfd[0].revents & (POLLERR | POLLNVAL))
		logerr("bad fd in poll set");

	if (pfd[0].revents & (POLLIN | POLLHUP)) {
		ret = readnet(netfd, buf, sizeof(uint64_t));
		if (ret <= 0 || ret == BADPKT)
			return -1;
		reply = unpack64(buf);
		if (n + 1 == reply)
			return 0;
	}
	return -1;
}

int
response(int netfd)
{
	unsigned char buf[MAXPAYLOADLEN];
	uint64_t reply;
	int ret;

	ret = readnet(netfd, buf, sizeof(uint64_t));
	if (ret <= 0 || ret == BADPKT)
		return -1;
	reply = unpack64(buf);
	pack64(buf, reply + 1);
	if (writenet(netfd, buf, sizeof(uint64_t)) <= 0)
		return -1;
	return 0;
}

int
tunnel(int netfd, int devfd)
{
	unsigned char buf[MAXPAYLOADLEN];
	struct pollfd pfd[2];
	int n;

	pfd[0].fd = netfd;
	pfd[0].events = POLLIN;
	pfd[1].fd = devfd;
	pfd[1].events = POLLIN;
	for (;;) {
		if (poll(pfd, 2, -1) < 0)
			logerr("poll failed");

		if (pfd[0].revents & (POLLERR | POLLNVAL) ||
		    pfd[1].revents & (POLLERR | POLLNVAL))
			logerr("bad fd in poll set");

		if (pfd[0].revents & (POLLIN | POLLHUP)) {
			n = readnet(netfd, buf, MAXPAYLOADLEN);
			if (n <= 0)
				return -1;
			if (n == BADPKT)
				logwarn("bad packet");
			else
				writedev(devfd, buf, n);
		}

		if (pfd[1].revents & (POLLIN | POLLHUP)) {
			n = readdev(devfd, buf, MAXPAYLOADLEN);
			if (n > 0) {
				n = writenet(netfd, buf, n);
				if (n <= 0)
					return -1;
			}
		}
	}
	return 0;
}

int
serversetup(int devfd)
{
	struct addrinfo hints, *ai, *p;
	struct sockaddr_in remote;
	struct timeval tv;
	int ret, netfd, listenfd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	ret = getaddrinfo(bindaddr, port, &hints, &ai);
	if (ret != 0)
		logerr("getaddrinfo: %s", gai_strerror(ret));

	for (p = ai; p; p = p->ai_next) {
		if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
			continue;
		setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (int []){1},
		           sizeof(int));
		if (bind(listenfd, p->ai_addr, p->ai_addrlen) < 0) {
			close(listenfd);
			continue;
		}
		break;
	}
	if (!p)
		logerr("failed to bind socket");
	freeaddrinfo(ai);

	if (listen(listenfd, 5) < 0)
		logerr("listen failed");

	revokeprivs();
	for (;;) {
		netfd = accept(listenfd, (struct sockaddr *)&remote,
		               (socklen_t []){sizeof(remote)});
		if (netfd < 0) {
			if (errno != ECONNABORTED)
				logwarn("accept failed");
			continue;
		}
		if (debug)
			logdbg("remote peer connected: %s",
			       inet_ntoa(remote.sin_addr));

		ms2tv(&tv, RCVTIMEO);
		setsockopt(netfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
		setsockopt(netfd, IPPROTO_TCP, TCP_NODELAY, (int []){1}, sizeof(int));

		if (challenge(netfd) < 0 || response(netfd) < 0) {
			logwarn("challenge-response failed");
			goto err;
		}

		tunnel(netfd, devfd);
err:
		close(netfd);
		if (debug)
			logdbg("remote peer disconnected: %s",
			       inet_ntoa(remote.sin_addr));
	}
}

int
clientsetup(int devfd)
{
	struct addrinfo hints, *ai, *p;
	struct timeval tv;
	int ret, netfd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	ret = getaddrinfo(host, port, &hints, &ai);
	if (ret != 0)
		logerr("getaddrinfo: %s", gai_strerror(ret));

	for (p = ai; p; p = p->ai_next) {
		if ((netfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
			continue;
		if (connect(netfd, p->ai_addr, p->ai_addrlen) < 0) {
			close(netfd);
			continue;
		}
		break;
	}
	freeaddrinfo(ai);
	if (!p) {
		close(netfd);
		logwarn("failed to connect to %s:%s", host, port);
		return -1;
	}
	if (debug)
		logdbg("connected to %s:%s", host, port);

	ms2tv(&tv, RCVTIMEO);
	setsockopt(netfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
	setsockopt(netfd, IPPROTO_TCP, TCP_NODELAY, (int []){1}, sizeof(int));

	if (response(netfd) < 0 || challenge(netfd) < 0) {
		logwarn("challenge-response failed");
		goto err;
	}

	tunnel(netfd, devfd);
err:
	close(netfd);
	logwarn("connection to %s:%s dropped", host, port);
	return -1;
}

void
usage(void)
{
	fprintf(stderr, "usage: stun [-df] -s [-b address] [-p port] [-t devtype] interface\n");
	fprintf(stderr, "       stun [-df] -h host [-p port] [-t devtype] interface\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	char *arg, *pw;
	int devfd;

	ARGBEGIN {
	case 'd':
		debug = 1;
		break;
	case 'f':
		foreground = 1;
		break;
	case 's':
		sflag = 1;
		break;
	case 'b':
		bindaddr = EARGF(usage());
		break;
	case 'h':
		host = EARGF(usage());
		break;
	case 'p':
		port = EARGF(usage());
		break;
	case 't':
		arg = EARGF(usage());
		if (!strcasecmp(arg, "tun"))
			devtype = TUNDEV;
		else if (!strcasecmp(arg, "tap"))
			devtype = TAPDEV;
		else
			usage();
		break;
	default:
		usage();
	} ARGEND

	if (argc != 1 || !(sflag ^ (host != NULL)))
		usage();

	signal(SIGPIPE, SIG_IGN);
	if (!foreground) {
		daemon(0, 0);
		openlog("stun", LOG_PID | LOG_NDELAY, LOG_DAEMON);
	}

	devfd = opendev(argv[0]);

	if (!(pw = getenv("STUNPW")))
		logerr("STUNPW is not set");
	derivekey(pw, strlen(pw));
	explicit_bzero(pw, strlen(pw));
	aesinit(&ectx, &dctx);

	if (sflag)
		return serversetup(devfd);
	revokeprivs();
	/* auto-reconnect client */
	for (;;) {
		clientsetup(devfd);
		sleep(RECONNECTTIMEO);
	}
	return 0;
}
